from .flower import Flower
from .flower import get_flower
from .flower import run_flower

from .task import FlowerTask


__all__ = [
    'Flower',
    'get_flower',
    'run_flower',
    'FlowerTask',
]

__version__ = '1.1.0'



import asyncio
from dataclasses import dataclass
from typing import Dict, Union


@dataclass
class TaskResult:
    task: asyncio.Task
    priority: int


class FlowerResult:
    """
    Main class for FlowerResult.
    Store the Flower task results and provide methods to access the results.
    """

    def __init__(self):
        self._result: Dict[str, TaskResult] = {}

    def add_result(self, task: asyncio.Task, task_id: str, priority: int):
        """
        Add a task to the result dict.

        :param task_id: Task id
        :param priority: Task sort priority
        :param task: Task to add
        """
        self._result[task_id] = TaskResult(task, priority)

    def all_result(self) -> Dict[str, TaskResult]:
        """
        Get all flower task result object.

        :return: task info.
        """
        return self._result

    def get_result(self, task_id: str) -> Union[TaskResult, None]:
        """
        Get flower task result object by task id.

        :param task_id: The ID of the task.
        :return: task info.
        """
        return self._result.get(task_id, None)

    def clear(self):
        """
        Clear flower result state.
        """
        self._result.clear()



import asyncio
from asyncio import Queue
from typing import Callable, Coroutine, Any


class AsyncProductionSystem:
    """
    Dev
    """
    def __init__(self, queue_size: int = 10):
        self.queue = Queue(maxsize=queue_size)
        self.producers = []
        self.consumers = []
        self.running = True

    async def start(
        self,
        num_producers: int,
        num_consumers: int,
        producer_fn: Callable[[], Coroutine[Any, Any, Any]],
        consumer_fn: Callable[[], Coroutine[Any, Any, Any]],
    ) -> None:
        self.running = True

        producer_Tasks = [
            asyncio.create_task(self._producer_loop(producer_fn))
            for _ in range(num_producers)
        ]

        consumer_Tasks = [
            asyncio.create_task(self._consumer_loop(consumer_fn))
            for _ in range(num_consumers)
        ]

        await asyncio.gather(*producer_Tasks, *consumer_Tasks)

    async def _producer_loop(self, producer_fn):
        while self.running:
            try:
                item = await producer_fn()
                await self.queue.put(item)
            except Exception as e:
                print(f"producer error: {e}")
                break

    async def _consumer_loop(self, consumer_fn):
        while self.running:
            try:
                item = await self.queue.get()
                await consumer_fn(item)
                self.queue.task_done()

            except Exception as e:
                print(f"consumer error: {e}")
                break

    def stop(self):
        self.running = False



from typing import Protocol, Dict, Any


class TaskCoroutine(Protocol):
    __qualname__: str

    async def __call__(self, ctx: Dict[Any, Any]) -> any: ...


class AsyncTaskCM(Protocol):
    __qualname__: str

    def __init__(self, **kwargs) -> None: ...

    async def __aenter__(self) -> None: ...

    async def __aexit__(self, exc_type, exc_value, traceback) -> None: ...


class SyncTaskCM(Protocol):
    __qualname__: str

    def __init__(self, **kwargs) -> None: ...

    def __enter__(self) -> None: ...

    def __exit__(self, exc_type, exc_value, traceback) -> None: ...



from typing import Any, Callable, Union, List

from .constant import DEFAULT_PRIORITY


class FlowerTask:
    """
    Main class for a task.
    """
    def __init__(self, *args: Any, **kwargs: Any):
        self.args = args
        self.kwargs = kwargs
        self.task_id = None
        self.function = None
        self.priority = DEFAULT_PRIORITY
        self.iter_keys = []

    def set_priority(self, priority: int) -> "FlowerTask":
        """
        Set priority for task.

        :param priority: The smaller the priority, the higher the priority to run.
        """
        self.priority = priority

        return self

    def set_function(self, function: Callable) -> "FlowerTask":
        """
        Set function for task.

        :param function: Function to be run.
        """
        self.function = function

        return self

    def set_task_id(self, task_id: str) -> "FlowerTask":
        """
        Set task_id for task.

        :param task_id: task_id for the task.
        """
        self.task_id = task_id

        return self

    def set_iter(self, iter_key: Union[str, List[str]]) -> "FlowerTask":
        """
        Set iter_key for task.

        :param iter_key: iter_key make task param iterable.
        """
        if isinstance(iter_key, str):
            self.iter_keys.append(iter_key)
        else:
            self.iter_keys.extend(iter_key)

        return self



from setuptools import setup, find_packages

setup(
    name='anf',
    version='1.1.0',
    description='Layout tasks with async and flower',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.12',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha',
    ],
    keywords='async flower task job',
    url='https://github.com/LL-Ling/anf.git',
    author='LL-Ling',
    author_email='1309399041@qq.com',
    license='MIT',
    packages=find_packages(),
    install_requires=[],
    include_package_data=False,
    zip_safe=False
)



import hashlib
import json
import re
import shutil

from .task import FlowerTask

import inspect
import itertools
from typing import List, Tuple, Dict, Iterable, Any, Optional, Callable
from inspect import Parameter


def get_function_params(func: callable) -> Dict[str, Dict]:
    """
    Get parameters of a function, including their properties.

    :param func: The function to inspect.
    :return: Dictionary with parameter names and their properties.
    """
    signature = inspect.signature(func)
    params = {}

    for name, param in signature.parameters.items():
        param_info = {
            'kind': param.kind,
            'default': param.default if param.default != Parameter.empty else None,
            'required': param.default == Parameter.empty and param.kind not in (Parameter.VAR_POSITIONAL,
                                                                                Parameter.VAR_KEYWORD)
        }
        params[name] = param_info
    return params


def permutate_task_params(task: FlowerTask, function: Optional[Callable] = None) -> List[Tuple[List, Dict]]:
    """
    Permutate a tasks parameters for a function.

    :param task: The FlowerTask task with parameters to permutate.
    :param function: The function of the task to permutate.
    :return: List for function parameters.
    """
    if not task.function and not function:
        raise ValueError("Task function is not set.")

    func = task.function or function
    sig = inspect.signature(func)
    params = get_function_params(func)

    try:
        bound_args = sig.bind(*task.args, **task.kwargs)
    except TypeError as e:
        raise RuntimeError(f"Invalid arguments for {func}: {e}") from e
    bound_args.apply_defaults()

    for key in task.iter_keys:
        if key not in params:
            raise RuntimeError(f"iter_key '{key}' does not exist in function parameters.")
        if params[key]['kind'] in (Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD):
            raise RuntimeError(f"iter_key '{key}' cannot be a variable-length parameter (*args or **kwargs).")
        value = bound_args.arguments[key]
        if not isinstance(value, Iterable) or isinstance(value, (str, bytes)):
            raise RuntimeError(f"iter_key '{key}' value must be a non-string/bytes iterable.")

    param_names = list(sig.parameters.keys())
    param_value_lists = []
    for name in param_names:
        value = bound_args.arguments[name]

        if name in task.iter_keys:
            if not isinstance(value, Iterable) or isinstance(value, (str, bytes)):
                raise RuntimeError(f"iter_key '{name}' value must be a non-string/bytes iterable.")
            param_value_lists.append(list(value))
        else:
            param_value_lists.append([value])

    permutations = itertools.product(*param_value_lists)

    results = []
    for perm in permutations:
        args = []
        kwargs = {}
        var_args = []
        var_kwargs = {}

        for name, value in zip(param_names, perm):
            param_kind = params[name]['kind']

            if param_kind == Parameter.VAR_POSITIONAL:
                var_args.extend(value if isinstance(value, Iterable) else [value])
            elif param_kind == Parameter.VAR_KEYWORD:
                var_kwargs.update(value if isinstance(value, dict) else {})
            elif param_kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD):
                args.append(value)
            else:
                kwargs[name] = value

        final_args = args + var_args
        final_kwargs = {**kwargs, **var_kwargs}

        results.append((final_args, final_kwargs))

    return results


def modify_msg_with_terminal(msg: str, msg_head_width: int = 0, fill_char: str = " ") -> str:
    """
    Fill the remaining msg length with fill_char to complete one line based on the terminal width using padding.

    :param msg: need modify msg.
    :param msg_head_width: msg head width.
    :param fill_char: default fill char.
    :return: link string
    """
    term_width = shutil.get_terminal_size().columns
    max_first = max(term_width - msg_head_width, 0)

    first_part = msg[:max_first]
    filled_first = first_part.ljust(max_first, fill_char)
    lines = [filled_first]

    remaining = msg[max_first:]
    while remaining:
        chunk = remaining[:term_width]
        filled_chunk = chunk.ljust(term_width, fill_char)
        lines.append(filled_chunk)
        remaining = remaining[term_width:]

    return '\n'.join(lines)


def link_param(*args, max_length: int = 40, **kwargs) -> str:
    """
    To join parameters with an underscore '_' and return a string.

    :param args: link parameters.
    :param max_length: max string length to link, previous link in etc.
    :param kwargs: link keyword arguments.
    :return: link string
    """
    def process_value(value: Any) -> str:
        if isinstance(value, (str, int, float, bool)):
            return re.sub(r'[^a-zA-Z0-9_]', '_', str(value))
        else:
            try:
                serialized = json.dumps(value, sort_keys=True)
            except (TypeError, ValueError):
                serialized = repr(value)
            return hashlib.sha1(serialized.encode()).hexdigest()[:8]

    parts = []

    for arg in args:
        parts.append(process_value(arg))

    for k, v in kwargs.items():
        safe_key = re.sub(r'[^a-zA-Z0-9_]', '_', str(k))
        parts.append(safe_key)
        parts.append(process_value(v))

    param_str = '_'.join(parts)
    param_str = re.sub(r'[^a-zA-Z0-9_]', '_', param_str)

    if len(param_str) > max_length:
        truncate_hash = hashlib.sha1(param_str.encode()).hexdigest()[:8]
        truncate_at = max_length - len(truncate_hash) - 1
        param_str = f"{param_str[:truncate_at]}_{truncate_hash}_etc"

    return param_str



import asyncio
import shutil
from dataclasses import dataclass
import itertools
import time
from typing import Callable, Optional, Any, List, Dict, Union, Tuple, Generator, Sequence
from uuid import uuid4

from ._typing import TaskCoroutine, AsyncTaskCM, SyncTaskCM
from .constant import DEFAULT_MAX_TASKS
from .result import FlowerResult
from .task import FlowerTask
from .utils import permutate_task_params


@dataclass
class _Progress:
    in_progress: List[str]
    fail_progress: List[str]
    total: int = 0
    success: int = 0


class Flower:
    """
    Main class for Flower.

    :param function: global execution function.
    :param max_tasks: maximum number of tasks to run at a time.
           Define a max_tasks like:
           max_tasks = 5 # global task max_tasks is 5
                or
           max_tasks = {
               0: 5,     # priority level of 0 task max_tasks is 5
               1: 3      # priority level of 1 task max_tasks is 3
           }
    :param progress_bar: write tasks progress bar.
    :param detector: function to determine the Success or Failure of a Task.
    :param cancel_while_failed: cancel all tasks while high priority task had failed.
    :param on_startup: coroutine function to run on task start.
    :param on_shutdown: coroutine function to run on task end.
    :param on_task_start: task function to run on task start.
    :param on_task_end: task function to run on task end.
    :param ctx: dict context.
    """
    def __init__(
        self,
        function: Optional[Callable] = None,
        max_tasks: Union[int, Dict[int, int]] = DEFAULT_MAX_TASKS,
        progress_bar: Optional[bool] = False,
        detector: Optional[Callable] = lambda x: not isinstance(x, Exception),
        cancel_while_failed: Optional[bool] = False,
        on_startup: Optional[List[TaskCoroutine]] = None,
        on_shutdown: Optional[List[TaskCoroutine]] = None,
        on_task_start: Optional[List[TaskCoroutine]] = None,
        on_task_end: Optional[List[TaskCoroutine]] = None,
        with_async_task: Optional[Union[AsyncTaskCM, Generator, Callable]] = None,
        with_sync_task: Optional[Union[SyncTaskCM, Generator, Callable]] = None,
        ctx: Optional[Dict[Any, Any]] = None,
    ):
        self.function = function
        self.cancel_while_failed = cancel_while_failed
        self.on_startup = on_startup or []
        self.on_shutdown = on_shutdown or []
        self.on_task_start = on_task_start or []
        self.on_task_end = on_task_end or []
        self.ctx = ctx or {}
        self.results = FlowerResult()
        self.detector = detector
        self._progress_bar = progress_bar
        self._with_async_task = with_async_task
        self._with_sync_task = with_sync_task
        self._max_tasks = max_tasks
        self._semaphore = None
        self._progress = _Progress(in_progress=[], fail_progress=[])
        self._pending_tasks: Dict[int, List[FlowerTask]] = {}

        if self._with_async_task and self._with_sync_task:
            raise ValueError("Only one of 'with_async_task' or 'with_sync_task' can be provided.")

    def __bool__(self) -> bool:
        return self._progress.total == self._progress.success

    def __repr__(self) -> str:
        pending_count = sum(len(tasks) for tasks in self._pending_tasks.values())
        return f'<{self.__class__.__name__} Pending: {pending_count}>'

    async def start_tasks(self) -> "Flower":
        """
        Start flower all tasks.

        :return: Flower obj
        """
        if not self._pending_tasks:
            raise RuntimeError(f'{self.__repr__()} has not no pending_tasks.')

        self.results.clear()
        self._progress = _Progress(in_progress=[], fail_progress=[])

        for event in self.on_startup:
            await event(self.ctx)

        for priority, flower_tasks in sorted(self._pending_tasks.items(), key=lambda x: x[0], reverse=False):
            self._define_semaphore(priority)
            tasks = list(itertools.chain(*[self._create_tasks(task) for task in flower_tasks]))

            if self._progress_bar:
                self._write_progress()

            if self._should_cancel_tasks(await asyncio.gather(*tasks, return_exceptions=True)):
                break

        for event in self.on_shutdown:
            await event(self.ctx)

        self._pending_tasks.clear()
        return self

    def add_task(self, flower_task: FlowerTask, _task_id: Optional[str] = None):
        """
        Add a task to the pending task list.

        :param flower_task: FlowerTask obj.
        :param _task_id: Custom FlowerTask obj task_id.
        """
        flower_task.task_id = _task_id or uuid4().hex[:15]
        self._pending_tasks.setdefault(flower_task.priority, []).append(flower_task)

    def progress_msg(self, max_show_progress: Union[int, None] = None, finish: bool = False) -> str:
        """
        Add a task to the pending task list.
            [ 100% ] | T/S/F: 200/200/0 | Progress Info

        :param max_show_progress: The maximum number of tasks to show progress bar for.
        :param finish: Return finish Progress.
        :return: msg string
        """
        failed_progress_count = len(self._progress.fail_progress)
        msg = (f"[ {((self._progress.success + failed_progress_count) / self._progress.total * 100):.0f}% ] | "
               f"T/S/F: {self._progress.total}/{self._progress.success}/{failed_progress_count} |")

        if finish:
            if max_show_progress and isinstance(max_show_progress, int):
                failed_progress = self._progress.fail_progress[-max_show_progress:]
            else:
                failed_progress = self._progress.fail_progress

            return f"{msg} FailProgress: {failed_progress}"
        else:
            if max_show_progress and isinstance(max_show_progress, int):
                in_progress = self._progress.in_progress[:max_show_progress]
            else:
                in_progress = self._progress.in_progress

            return f"{msg} InProgress: {in_progress}"

    def _define_semaphore(self, priority: int):
        """
        Define a semaphore based on the priority and max_tasks configuration.

        :param priority: The priority level of the task.
        """
        max_tasks = self._max_tasks.get(priority,
                                        DEFAULT_MAX_TASKS) if isinstance(self._max_tasks, dict) else self._max_tasks
        self._semaphore = asyncio.Semaphore(max_tasks)

    def _create_tasks(self, flower_task: FlowerTask) -> List[asyncio.Task]:
        """
        Create a coroutine_task and add it to the tasks list.

        :param flower_task: FlowerTask obj.
        :return: A coroutine task list.
        """
        if flower_task.function is None and self.function is None:
            raise RuntimeError(f'priority {flower_task.priority} exists function-defining task.')

        if flower_task.function is None:
            flower_task.function = self.function

        task_params = permutate_task_params(flower_task)
        if not task_params:
            raise RuntimeError(f'priority {flower_task.function} no exists any params.')

        tasks = []
        for index, (args, kwargs) in enumerate(task_params):
            task_id = f'{flower_task.task_id}-{index}' if flower_task.iter_keys else flower_task.task_id

            ctx = {
                "task_id": task_id,
                "function_name": flower_task.function.__name__,
                "priority": flower_task.priority,
                "args": args,
                "kwargs": kwargs,
                "start_time": 0,
                "end_time": 0,
                "result": None
            }
            ctx = {**self.ctx, **ctx}

            task = asyncio.create_task(self._wrap_semaphore(flower_task.function, args, kwargs, ctx=ctx))
            tasks.append(task)

            if self._progress_bar:
                task.add_done_callback(lambda _: self._write_progress())

        self._progress.total += len(tasks)
        return tasks

    def _should_cancel_tasks(self, results: Tuple[Union[Any, Exception]]) -> bool:
        """
        Determine if tasks should be canceled based on results.

        :param results: List of task results.
        :return: True if tasks should be cancelled, otherwise False.
        """
        if not self.cancel_while_failed:
            return False

        for res in results:
            if not self.detector(res):
                return True

        return False

    def _write_progress(self):
        """
        Write progress bar to progress bar if in_progress have any tasks.
        """
        if self._progress_bar:
            def get_truncated_message(message: str):
                return message[:max(shutil.get_terminal_size().columns - 1, 0)]

            msg = get_truncated_message(self.progress_msg())
            print(f"{msg}", end="\r")

    async def _wrap_semaphore(self, function: Callable, args: Any, kwargs: Any, ctx: Dict[Any, Any]) -> Any:
        """
        Make task work semaphore away.

        :param function: task function
        :return: coroutine result
        """
        async with self._semaphore:
            ctx["start_time"] = time.time()

            for event in self.on_task_start:
                await event(ctx)

            try:
                if self._with_sync_task:
                    with self._with_sync_task(**ctx):
                        task = asyncio.create_task(function(*args, **kwargs))
                elif self._with_async_task:
                    async with self._with_async_task(**ctx):
                        task = asyncio.create_task(function(*args, **kwargs))
                else:
                    task = asyncio.create_task(function(*args, **kwargs))

                self._progress.in_progress.append(ctx["task_id"])
                self.results.add_result(task, ctx["task_id"], ctx["priority"])
                res = await asyncio.wait_for(task, timeout=None)
                ctx["result"] = res
                if self.detector(ctx["result"]):
                    self._progress.success += 1
                else:
                    self._progress.fail_progress.append(ctx["task_id"])
            except Exception as exc:
                res = exc
                ctx["result"] = exc
                ctx["stack"] = task.get_stack() if task else []
                self._progress.fail_progress.append(ctx["task_id"])
            finally:
                ctx["end_time"] = time.time()
                self._progress.in_progress.remove(ctx["task_id"])

            for event in self.on_task_end:
                await event(ctx)

            return res


def get_flower(function: Callable, **kwargs: Any) -> Flower:
    """
    Get flower object for given function.

    :param function: Task function.
    :param kwargs: Flower kwargs
    :return: Flower obj
    """
    return Flower(function, **kwargs)


async def run_flower(function_tasks: Sequence[Any], function: Optional[Callable] = None, **kwargs: Any) -> Flower:
    """
    Run simple task.

    :param function_tasks: Iterable param for function.
    :param function: Run with function.
    :param kwargs: Flower kwargs
    :return: Flower obj
    """
    flower = get_flower(function, **kwargs)

    for task in function_tasks:
        flower_task = task if isinstance(task, FlowerTask) else FlowerTask(task)
        flower.add_task(flower_task, _task_id=flower_task.task_id)

    return await flower.start_tasks()
